import uuid
from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from django_extensions.db.models import TimeStampedModel

from bank.managers import TransferManager, FreezeTransferManager
from user.models import User
# from kuailian.core.storage_backends import PrivateMediaRootS3Boto3Storage
# from kuailian.bank.managers import UserBankProfileManager

__all__ = [
    'MarkupFee',
    'PersonalBankAccount',
    'Transfer',
    # 'KuailianBankAccount',
    # 'GlobalBankLimit',
    'PersonalBankLimit',
    # 'BankInvoice',
    'UserWallet',
    'DestinationPayment',
    # 'BankProduct',
    # 'DefaultBankProductUser',
    # 'FreezeTransfer',
    # 'FinMonitoringLog'
]


class MarkupFee(TimeStampedModel):
    SELL_INVISIBLE_FEE = 'sell_invisible_fee'
    SELL_VISIBLE_FEE = 'sell_visible_fee'
    BUY_INVISIBLE_FEE = 'buy_invisible_fee'
    BUY_VISIBLE_FEE = 'buy_visible_fee'
    EUROPEAN_TRANSFER_FEE = 'european_transfer_fee'
    INTERNATIONAL_TRANSFER_FEE = 'international_transfer_fee'

    TYPES = (
        (SELL_INVISIBLE_FEE, _('Sell Invisible Fee')),
        (SELL_VISIBLE_FEE, _('Sell Visible Fee')),
        (BUY_INVISIBLE_FEE, _('Buy Invisible Fee')),
        (BUY_VISIBLE_FEE, _('Buy Visible Fee')),
        (EUROPEAN_TRANSFER_FEE, _('European Transfer Fee')),
        (INTERNATIONAL_TRANSFER_FEE, _('International Transfer Fee'))
    )
    type = models.CharField(
        unique=True,
        max_length=100,
        choices=TYPES
    )
    percent_value = models.DecimalField(
        verbose_name=_('Percent Value'),
        max_digits=4,
        decimal_places=2,
    )
    fixed_value = models.DecimalField(
        verbose_name=_('Fixed Value'),
        max_digits=4,
        decimal_places=2,
    )
    description = models.TextField(
        max_length=256,
        null=True,
        blank=True
    )


class PersonalBankAccount(TimeStampedModel):
    NOT_VERIFIED = 'not_verified'
    PENDING = 'pending'
    VERIFIED = 'verified'
    DECLINED = 'declined'
    STATUS_CHOICES = (
        (NOT_VERIFIED, _('Not Verified')),
        (PENDING, _('Pending')),
        (VERIFIED, _('Verified')),
        (DECLINED, _('Declined'))
    )
    status = models.CharField(
        max_length=16,
        choices=STATUS_CHOICES,
        default=NOT_VERIFIED
    )
    iban = models.CharField(
        max_length=100,
        null=False,
        blank=False
    )
    swift = models.CharField(
        max_length=100,
        null=False,
        blank=False
    )
    recipient_name = models.CharField(
        max_length=255,
        null=False,
        blank=False
    )
    bank_name = models.CharField(
        max_length=255,
        null=False,
        blank=False
    )
    bank_country = models.CharField(
        max_length=255,
        null=False,
        blank=False
    )
    recipient_country = models.CharField(
        max_length=255,
        null=False,
        blank=False
    )
    physical_address = models.CharField(
        max_length=255,
        null=False,
        blank=False
    )
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='bank_accounts',
    )
    deleted_date = models.DateTimeField(
        _('Deleted Date'),
        null=True,
        blank=True,
    )
    verified_date = models.DateTimeField(
        _('Verified Date'),
        null=True,
        blank=True,
    )

    @property
    def is_european(self):
        from bank.helpers import is_europen_bank
        return is_europen_bank(self)

    @property
    def service_cost_fixed(self):
        markup_fee = self.get_makup_fee()
        return markup_fee.fixed_value

    @property
    def service_cost_percent(self):
        markup_fee = self.get_makup_fee()
        return markup_fee.percent_value

    def get_makup_fee(self):
        transfer_type = MarkupFee.EUROPEAN_TRANSFER_FEE if self.is_european else MarkupFee.INTERNATIONAL_TRANSFER_FEE
        markup_fee = MarkupFee.objects.get(type=transfer_type)
        return markup_fee

    class Meta:
        verbose_name = "Personal chosschange accounts"
        verbose_name_plural = "Personal chosschange accounts"


class Transfer(TimeStampedModel):
    SELL = 'sell'
    BUY = 'buy'
    VERIFY = 'verify'
    WITHDRAW = 'withdraw'
    DEPOSIT = 'deposit'
    TRANSFER_TYPES = (
        (SELL, _('Sell')),
        (BUY, _('Buy')),
        (VERIFY, _('Verify')),
        (WITHDRAW, _('Withdraw')),
        (DEPOSIT, _('Deposit')),
    )

    transfer_type = models.CharField(
        max_length=10,
        choices=TRANSFER_TYPES,
        verbose_name=_('Transaction Type')
    )

    # TODO: what difference between new and pending? Can we merge them?
    NEW = 'new'
    PENDING = 'pending'
    IN_PROGRESS = 'in_progress'
    COMPLETED = 'completed'
    CANCELED = 'canceled'
    REFUNDED = 'refunded'
    ON_HOLD = 'on_hold'

    STATUS_CHOICES = (
        (NEW, _('New')),
        (IN_PROGRESS, _('In Progress')),
        (PENDING, _('Pending')),
        (COMPLETED, _('Completed')),
        (CANCELED, _('Canceled')),
        (REFUNDED, _('Refunded')),
        (ON_HOLD, _('On Hold')),
    )
    KRAKEN_STATUS_OPEN = 'open'
    KRAKEN_STATUS_PENDING = 'pending'
    KRAKEN_STATUS_CLOSED = 'closed'
    KRAKEN_STATUS_CANCELED = 'canceled'
    KRAKEN_STATUS_EXPIRED = 'expired'
    # our custom status
    KRAKEN_STATUS_UNKNOWN = 'unknown'

    WITHDRAW_STATUS_INITIATED = 'Initial'
    WITHDRAW_STATUS_PENDING = 'Pending'
    WITHDRAW_STATUS_SENDING = 'Sending'
    WITHDRAW_STATUS_SUCCESS = 'Success'
    WITHDRAW_STATUS_ON_HOLD = 'On Hold'
    WITHDRAW_STATUS_FAILURE = 'Failure'
    WITHDRAW_STATUS_SETTLED = 'Settled'
    # our custom status
    WITHDRAW_STATUS_UNKNOWN = 'Unknown'

    WITHDRAW_STATUS_CHOICES = (
        (WITHDRAW_STATUS_INITIATED, _('Initiated')),
        (WITHDRAW_STATUS_PENDING, _('Pending')),
        (WITHDRAW_STATUS_SENDING, _('Sending')),
        (WITHDRAW_STATUS_SUCCESS, _('Success')),
        (WITHDRAW_STATUS_ON_HOLD, _('On Hold')),
        (WITHDRAW_STATUS_FAILURE, _('Failure')),
        (WITHDRAW_STATUS_SETTLED, _('Settled')),
        (WITHDRAW_STATUS_UNKNOWN, _('Unknown')),
    )

    KRAKEN_STATUS_CHOICES = (
        (KRAKEN_STATUS_PENDING, _('Pending')),
        (KRAKEN_STATUS_OPEN, _('Open')),
        (KRAKEN_STATUS_CLOSED, _('Closed')),
        (KRAKEN_STATUS_CANCELED, _('Canceled')),
        (KRAKEN_STATUS_EXPIRED, _('Expired')),
        (KRAKEN_STATUS_UNKNOWN, _('Unknown')),
    )

    status = models.CharField(
        max_length=16,
        choices=STATUS_CHOICES,
        default=NEW,
        verbose_name=_('Transaction Status'),
        help_text=_("Note: Status Refunded can not be changed")
    )

    EUR = 'eur'
    ETH = 'eth'
    XRP = 'xrp'
    BTC = 'btc'
    LTC = 'ltc'
    DASH = 'dash'
    USDT = 'usdt'
    USDC = 'usdc'
    BCH = 'bch'
    XTZ = 'xtz'

    CRYPTOCURRENCY_TYPES = (
        (EUR, 'Euro'),
        (ETH, 'Ethereum'),
        (XRP, 'Ripple'),
        (BTC, 'Bitcoin'),
        (LTC, 'Litecoin'),
        (BCH, 'Bitcoin Cash'),
        (DASH, 'Dash'),
        (USDT, 'Tether'),
        (USDC, 'USD Coin'),
        (XTZ, 'Tezos')
    )
    cryptocurrency = models.CharField(
        max_length=9,
        null=False,
        blank=False,
        choices=CRYPTOCURRENCY_TYPES,
        verbose_name=_('Currency')
    )

    BANK_ACCOUNT = 'bank_account'
    KUAILIAN_BANK_ACCOUNT = 'kuailian_bank_account'

    ACCOUNT_TYPES = (
        (BANK_ACCOUNT, _('Bank Account')),
        (KUAILIAN_BANK_ACCOUNT, _('Crosschange Bank Account'))
    )
    account_type = models.CharField(
        null=False,
        blank=False,
        max_length=32,
        choices=ACCOUNT_TYPES,
        default=KUAILIAN_BANK_ACCOUNT
    )
    iban = models.CharField(
        max_length=100,
        blank=True
    )
    swift = models.CharField(
        max_length=100,
        blank=True
    )
    recipient_name = models.CharField(
        max_length=255,
        blank=True
    )
    bank_name = models.CharField(
        max_length=255,
        blank=True
    )
    bank_country = models.CharField(
        max_length=255,
        blank=True
    )
    recipient_country = models.CharField(
        max_length=255,
        blank=True
    )
    physical_address = models.CharField(
        max_length=255,
        blank=True
    )
    exchange_rate = models.ForeignKey(
        'invoices.EthExchangeRate',
        on_delete=models.CASCADE,
        null=True,
        blank=True
    )
    exchange_rate_value = models.DecimalField(
        max_digits=23,
        decimal_places=18,
        null=True,
        blank=True
    )
    applied_invisible_fee = models.DecimalField(
        max_digits=100,
        decimal_places=18,
        null=True,
        blank=True,
        verbose_name=_('Invisible fee')
    )

    withdraw_fee = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0,
        verbose_name=_('Withdraw fee')
    )
    applied_visible_fee = models.DecimalField(
        max_digits=100,
        decimal_places=18,
        null=True,
        blank=True,
        verbose_name=_('Visible fee')
    )
    original_pay_amount = models.DecimalField(
        max_digits=100,
        decimal_places=18,
        null=True,
        blank=True,
    )
    pay_amount = models.DecimalField(
        max_digits=100,
        decimal_places=18
    )
    receive_amount = models.DecimalField(
        max_digits=100,
        decimal_places=18
    )
    wallet_address = models.CharField(
        max_length=100,
        blank=True,
    )
    eth_tx = models.CharField(
        _('Ethereum transaction'),
        max_length=66,
        blank=True
    )
    user = models.ForeignKey(
        User,
        related_name='transfers_ids',
        on_delete=models.CASCADE)

    # TODO: remove null=True
    additional_details = models.TextField(
        max_length=256,
        null=True,
        blank=True
    )
    completed_date = models.DateTimeField(
        _('Completed Date'),
        auto_now_add=False,
        null=True,
        blank=True,
    )
    canceled_date = models.DateTimeField(
        _('Canceled Date'),
        auto_now_add=False,
        null=True,
        blank=True,
    )
    kb_commission_calculated = models.BooleanField(default=False)
    personal_bank_account = models.ForeignKey(
        PersonalBankAccount,
        on_delete=models.CASCADE,
        null=True,
        blank=False
    )
    # TODO: remove fields after merge https://oua.atlassian.net/browse/KBANK-296
    kraken_order_txid = models.CharField(
        max_length=255,
        blank=True
    )
    kraken_withdraw_refid = models.CharField(
        max_length=255,
        blank=True
    )
    kraken_status = models.CharField(
        max_length=32,
        choices=KRAKEN_STATUS_CHOICES,
        blank=True
    )
    kraken_withdraw_status = models.CharField(
        max_length=32,
        choices=WITHDRAW_STATUS_CHOICES,
        blank=True
    )
    kraken_withdraw_status_fee = models.CharField(
        max_length=32,
        choices=WITHDRAW_STATUS_CHOICES,
        blank=True
    )
    kraken_order_txid_for_fee = models.CharField(
        max_length=255,
        blank=True
    )
    kraken_withdraw_refid_for_fee = models.CharField(
        max_length=255,
        blank=True
    )
    sender_tag = models.CharField(
        max_length=32,
        blank=True,
        help_text=_('Sender tags uses for sending Ripple to identify sender')
    )
    objects = TransferManager()

    class Meta:
        verbose_name = _('Transaction')
        verbose_name_plural = _('Transactions')

    def __str__(self):
        return f'Date: ' + str(self.completed_date) + ', Receive Amount: ' + str(self.receive_amount)
        # return f'Invoice: {self.eth_wallet}'

    def get_last_unprocessed_payment(self):
        return DestinationPayment.objects.filter(
            transfer=self, is_processed=False
        ).order_by('created').first()


# class KuailianBankAccount(TimeStampedModel):
#     iban = models.CharField(
#         max_length=100,
#         null=False,
#         blank=False
#     )
#     swift = models.CharField(
#         max_length=100,
#         null=False,
#         blank=False
#     )
#     recipient_name = models.CharField(
#         max_length=255,
#         null=False,
#         blank=False
#     )
#     bank_name = models.CharField(
#         max_length=255,
#         null=False,
#         blank=False
#     )
#     bank_country = models.CharField(
#         max_length=255,
#         null=False,
#         blank=False
#     )
#     recipient_country = models.CharField(
#         max_length=255,
#         null=False,
#         blank=False
#     )
#     physical_address = models.CharField(
#         max_length=255,
#         null=False,
#         blank=False
#     )
#     is_active = models.BooleanField(default=False)

#     def save(self, *args, **kwargs):
#         if self.is_active:
#             qs = type(self).objects.filter(is_active=True)
#             if self.pk:
#                 qs = qs.exclude(pk=self.pk)
#             qs.update(is_active=False)

#         super(KuailianBankAccount, self).save(*args, **kwargs)

#     class Meta:
#         verbose_name = "Crosschange Bank Account"
#         verbose_name_plural = "Crosschange Bank Accounts"


# class GlobalBankLimit(TimeStampedModel):
#     SELL = 'sell'
#     BUY = 'buy'
#     WITHDRAW = 'withdraw'
#     TRANSFER_TYPES = (
#         (SELL, _('Sell')),
#         (BUY, _('Buy')),
#         (WITHDRAW, _('Withdraw')),
#     )
#     transfer_type = models.CharField(
#         max_length=10,
#         choices=TRANSFER_TYPES,
#         verbose_name=_('Transaction Type'),
#         unique=True,
#     )
#     min_limit = models.DecimalField(
#         max_digits=100,
#         decimal_places=2
#     )
#     max_limit = models.DecimalField(
#         max_digits=100,
#         decimal_places=2
#     )


class PersonalBankLimit(TimeStampedModel):
    SELL = 'sell'
    BUY = 'buy'
    WITHDRAW = 'withdraw'
    TRANSFER_TYPES = (
        (SELL, _('Sell')),
        (BUY, _('Buy')),
        (WITHDRAW, _('Withdraw')),
    )

    APPROVED = 'approved'
    DECLINED = 'declined'
    PENDING = 'pending'
    LIMIT_STATUS_CHOICES = (
        (APPROVED, _('Approved')),
        (DECLINED, _('Declined')),
        (PENDING, _('Pending')),
    )
    limit_status = models.CharField(
        max_length=20,
        choices=LIMIT_STATUS_CHOICES,
        verbose_name=_('Status'),
    )
    transfer_type = models.CharField(
        max_length=10,
        choices=TRANSFER_TYPES,
        verbose_name=_('Transaction Type'),
    )
    min_limit = models.DecimalField(
        max_digits=100,
        decimal_places=2
    )
    max_limit = models.DecimalField(
        max_digits=100,
        decimal_places=2
    )
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE
    )
    approved_date = models.DateTimeField(
        _('Approved Date'),
        auto_now_add=False,
        null=True,
        blank=True,
    )
    declined_date = models.DateTimeField(
        _('Declined Date'),
        auto_now_add=False,
        null=True,
        blank=True,
    )

    class Meta:
        verbose_name = "Personal crosschange limit"
        verbose_name_plural = "Personal crosschange limits"


# def upload_to(instance, filename):
#     return 'bank_invoices/{0}/{1}'.format(instance.user.id, filename)


# class BankInvoice(TimeStampedModel):
#     SELL = 'sell'
#     BUY = 'buy'
#     VERIFY = 'verify'
#     WITHDRAW = 'withdraw'
#     DEPOSIT = 'deposit'
#     TRANSFER_TYPES = (
#         (SELL, _('Sell')),
#         (BUY, _('Buy')),
#         (VERIFY, _('Verify')),
#         (WITHDRAW, _('Withdraw')),
#         (DEPOSIT, _('Deposit')),
#     )

#     transfer_type = models.CharField(
#         max_length=10,
#         choices=TRANSFER_TYPES,
#         verbose_name=_('Transaction Type')
#     )
#     uuid = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
#     date_issued = models.DateField(blank=True)
#     user = models.ForeignKey(User, on_delete=models.PROTECT)
#     full_name = models.CharField(max_length=255, blank=True)
#     username = models.CharField(max_length=150, blank=True)
#     email = models.EmailField(blank=True)
#     phone = models.CharField(max_length=64, blank=True)
#     reference_number = models.CharField(_('Reference number'), max_length=24, blank=True)
#     sum_amount_eth = models.DecimalField(max_digits=23, decimal_places=18, default=0)
#     sum_amount_eur = models.DecimalField(max_digits=10, decimal_places=2, default=0)
#     sum_amount_usd = models.DecimalField(max_digits=10, decimal_places=2, default=0)
#     transfer = models.OneToOneField(Transfer, on_delete=models.PROTECT)

#     pdf = models.FileField(
#         upload_to=upload_to, blank=True, null=True,
#         storage=PrivateMediaRootS3Boto3Storage() if hasattr(settings, 'AWS_PRIVATE_STORAGE_BUCKET_NAME') else None
#     )

#     class Meta:
#         verbose_name = "Crosschange Invoice"
#         verbose_name_plural = "Crosschange Invoice"


class UserWallet(models.Model):
    ETH = 'eth'
    XRP = 'xrp'
    BTC = 'btc'
    LTC = 'ltc'
    DASH = 'dash'
    USDT = 'usdt'
    USDC = 'usdc'
    BCH = 'bch'
    XTZ = 'xtz'

    CRYPTOCURRENCY_TYPES = (
        (ETH, 'Ethereum'),
        (XRP, 'Ripple'),
        (BTC, 'Bitcoin'),
        (LTC, 'Litecoin'),
        (BCH, 'Bitcoin Cash'),
        (DASH, 'Dash'),
        (USDT, 'Tether'),
        (USDC, 'USD Coin'),
        (XTZ, 'Tezos')
    )
    cryptocurrency = models.CharField(
        max_length=9,
        null=False,
        blank=False,
        choices=CRYPTOCURRENCY_TYPES,
        verbose_name='Currency'
    )
    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='wallets'
    )
    address = models.CharField('Address', max_length=100)

    class Meta:
        verbose_name = 'user wallet'
        verbose_name_plural = 'Users Wallets'

    def __str__(self):
        return self.address


class DestinationPayment(TimeStampedModel):
    XRP = 'xrp'
    BTC = 'btc'
    BCH = 'bch'
    LTC = 'ltc'
    DASH = 'dash'
    CRYPTOCURRENCY_TYPES = (
        (XRP, 'Ripple'),
        (BTC, 'Bitcoin'),
        (BCH, 'Bitcoin Cash'),
        (LTC, 'Litecoin'),
        (DASH, 'Dash')
    )
    cryptocurrency = models.CharField(
        max_length=9,
        null=False,
        blank=False,
        choices=CRYPTOCURRENCY_TYPES,
        verbose_name='Currency'
    )
    transfer = models.ForeignKey(
        Transfer,
        blank=True,
        null=True,
        on_delete=models.PROTECT
    )
    amount = models.DecimalField(
        max_digits=100,
        decimal_places=18,
        null=True,
        blank=True,
        verbose_name='Destination payment amount'
    )
    tx_hash = models.CharField(
        'Transaction hash',
        max_length=66,
        blank=True,
        unique=True
    )
    is_processed = models.BooleanField(default=False)


# class BankProduct(models.Model):
#     KEY_BANK = 'bank'

#     PROJECT_KEYS = (
#         (KEY_BANK, 'Bank'),
#     )
#     title = models.CharField(max_length=100)
#     project_key = models.CharField(max_length=20, choices=PROJECT_KEYS)
#     short_description = models.CharField(max_length=200)
#     is_activation_fee_required = models.BooleanField()
#     is_kyc_required = models.BooleanField()

#     class Meta:
#         verbose_name = "Crosschange Product"
#         verbose_name_plural = "Crosschange Product"


# class DefaultBankProductUser(models.Model):
#     user = models.OneToOneField(User, on_delete=models.PROTECT)
#     product = models.ForeignKey(BankProduct, on_delete=models.CASCADE)


# class FreezeTransfer(Transfer):
#     objects = FreezeTransferManager()

#     class Meta:
#         proxy = True
#         verbose_name = "Freeze Transfer"
#         verbose_name_plural = "Freeze Transfers"


# class FinMonitoringLog(models.Model):
#     transfer = models.ForeignKey(Transfer, on_delete=models.CASCADE)
#     condition = models.CharField(max_length=100, verbose_name="Condition name")
#     result = models.BooleanField(verbose_name="Condition result")
#     reason = models.CharField(max_length=2000, verbose_name="Reason")
#     created = models.DateTimeField(auto_created=True)

#     class Meta:
#         verbose_name = "Fin monitoring log"
#         verbose_name_plural = "Fin monitoring logs"
#         db_table = "fin_monitoring_logs"


# class UserBankProfile(models.Model):
#     user = models.OneToOneField(User, on_delete=models.CASCADE)
#     toc = models.DateTimeField(_("Crosschange Terms and Conditions"), null=True, default=None)

#     objects = UserBankProfileManager()

#     class Meta:
#         verbose_name = "Crosschange Profile"
#         verbose_name_plural = "Crosschange Profile"
