import logging
from decimal import Decimal
from typing import List, Dict, Union, Any
from django.conf import settings

from interfaces.bitcoin.helpers import (
    get_btc_wallet_balance, get_btc_received_by_payment_list,
    get_entire_btc_wallet_balance, is_btc_wallet_address
)
from interfaces.bitcoin_cash.helpers import (
    get_bch_wallet_balance, get_bch_received_by_payment_list,
    get_entire_bch_wallet_balance, is_bch_wallet_address
)
from interfaces.constants import (
    ETH, XRP, BTC, BCH, LTC, USDC, DASH, USDT, XTZ,
    ETHEREUM_BASED_CRYPTO, BITCOIN_BASED_CRYPTO
)
from interfaces.dash.helpers import (
    get_dash_wallet_balance, get_dash_received_by_payment_list,
    get_entire_dash_wallet_balance, is_dash_wallet_address
)
from interfaces.ethereum.explorer import EthExplorerApi
from interfaces.ethereum.helpers import (
    get_network_url, get_eth_wallet_balance, is_eth_wallet_address
)
from interfaces.explorer import BtxExplorerApi
from interfaces.litecoin.helpers import (
    get_ltc_wallet_balance, get_ltc_received_by_payment_list,
    get_entire_ltc_wallet_balance
)
from interfaces.litecoin.helpers import is_ltc_wallet_address
from interfaces.ripple.explorer import XrpExplorerApi
from interfaces.ripple.helpers import (
    get_xrp_wallet_balance, get_xrp_amount_and_hash,
    is_xrp_wallet_address
)
from interfaces.tether.helpers import (
    get_usdt_wallet_balance, is_tether_wallet_address
)
from interfaces.tezos.helpers import (
    get_tezos_wallet_balance, is_tezos_wallet_address
)
from interfaces.usdc.helpers import get_usdc_wallet_balance
from bank.constants import AVAILABLE_CRYPTOCURRENCIES
from bank.models import Transfer
from invoices.models import EthExchangeRate

__all__ = [
    'is_wallet_address', 'domain_tx_link', 'wallet_link',
    'get_balance', 'get_maximum_tx_fee', 'get_liquidity_address',
    'get_liquidity_balance', 'get_received_payment', 'get_liquidity_dict',
    'get_transactions'
]
logger = logging.getLogger(__name__)


def is_wallet_address(address: str, crypto: str = ETH) -> bool:
    """
    Helper method for wallet address validation
    """
    if crypto in (ETH, USDC):
        return is_eth_wallet_address(address)
    elif crypto == XRP:
        return is_xrp_wallet_address(address)
    elif crypto == BTC:
        return is_btc_wallet_address(address)
    elif crypto == BCH:
        return is_bch_wallet_address(address)
    elif crypto == LTC:
        return is_ltc_wallet_address(address)
    elif crypto == DASH:
        return is_dash_wallet_address(address)
    elif crypto == USDT:
        return is_tether_wallet_address(address)
    elif crypto == XTZ:
        return is_tezos_wallet_address(address)


def domain_tx_link(crypto: str = ETH, eth_tx: str = None) -> str:
    domain = ''
    if crypto == ETH or (crypto in [USDC, USDT] and eth_tx):
        if settings.GETH_NETWORK == 'rinkeby':
            network_url = 'rinkeby.'
        else:
            network_url = ''
        domain = f'{network_url}etherscan.io'
    elif crypto == XRP:
        domain = 'xrpscan.com'
    elif crypto == BTC:
        domain = 'live.blockcypher.com/btc'
    elif crypto == BCH:
        domain = 'explorer.bitcoin.com/bch'
    elif crypto == LTC:
        domain = 'live.blockcypher.com/ltc'
    elif crypto == DASH:
        domain = 'live.blockcypher.com/dash'
    elif crypto == USDC:
        domain = 'etherscan.io/token/0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48'
    elif crypto == USDT:
        domain = 'etherscan.io/token/0xdac17f958d2ee523a2206206994597c13d831ec7'
    elif crypto == XTZ:
        domain = 'tezblock.io'

    if eth_tx:
        return f'https://{domain}/tx/{eth_tx}'

    return domain


def wallet_link(address: str, crypto: str = ETH) -> str:
    """
    Helper method to retrieve url for wallet address
    """
    domain = domain_tx_link(crypto)
    url = ''
    if crypto == ETH:
        url = 'https://{}/address/{}'.format(
            domain, address
        )
    elif crypto == XRP:
        url = 'https://bithomp.com/explorer/{}'.format(address)
    elif crypto == BTC:
        url = 'https://{}/address/{}/'.format(
            domain, address
        )
    elif crypto == BCH:
        url = 'https://{}/address/{}/'.format(
            domain, address
        )
    elif crypto == LTC:
        url = 'https://{}/address/{}/'.format(
            domain, address
        )
    elif crypto == DASH:
        url = 'https://{}/address/{}/'.format(
            domain, address
        )
    elif crypto == USDC:
        url = 'https://{}?a={}'.format(
            domain, address
        )
    elif crypto == USDT:
        url = 'https://{}?a={}'.format(
            domain, address
        )
    elif crypto == XTZ:
        url = 'https://{}/account/{}'.format(
            domain, address
        )
    return url


def get_balance(address: str, crypto: str = ETH):
    if crypto == ETH:
        balance = get_eth_wallet_balance(address)
    elif crypto == XRP:
        balance = get_xrp_wallet_balance(address)
    elif crypto == BTC:
        balance = get_btc_wallet_balance(address)
    elif crypto == LTC:
        balance = get_ltc_wallet_balance(address)
    elif crypto == DASH:
        balance = get_dash_wallet_balance(address)
    elif crypto == USDT:
        balance = get_usdt_wallet_balance(address)
    elif crypto == USDC:
        balance = get_usdc_wallet_balance(address)
    elif crypto == BCH:
        balance = get_bch_wallet_balance(address)
    elif crypto == XTZ:
        balance = get_tezos_wallet_balance(address)
    else:
        raise NotImplementedError
    return balance


def get_maximum_tx_fee(cryptocurrency):
    """
    This values should be updated further
    """
    if cryptocurrency == ETH:
        fee = settings.KRAKEN_WITHDRAW_FEE_ETH
    elif cryptocurrency == BTC:
        fee = settings.KRAKEN_WITHDRAW_FEE_BTC
    elif cryptocurrency == LTC:
        fee = settings.KRAKEN_WITHDRAW_FEE_LTC
    elif cryptocurrency == XRP:
        fee = settings.KRAKEN_WITHDRAW_FEE_XRP
    elif cryptocurrency == BCH:
        fee = settings.KRAKEN_WITHDRAW_FEE_BCH
    elif cryptocurrency in USDC:
        eth_exchange_rate = EthExchangeRate.get_transaction_price_for_pair(
            pair=EthExchangeRate.USDC_ETH, transaction=Transfer.SELL
        )
        # USDC is a token so two ethereum transactions are required
        fee = Decimal(2 * settings.KRAKEN_WITHDRAW_FEE_ETH) / eth_exchange_rate
    elif cryptocurrency == USDT:
        eth_exchange_rate = EthExchangeRate.get_transaction_price_for_pair(
            pair=EthExchangeRate.USDT_ETH, transaction=Transfer.SELL
        )
        # USDT is a token so two ethereum transactions are required
        fee = Decimal(2 * settings.KRAKEN_WITHDRAW_FEE_ETH) / eth_exchange_rate
    elif cryptocurrency == DASH:
        fee = settings.KRAKEN_WITHDRAW_FEE_DASH
    elif cryptocurrency == XTZ:
        fee = settings.KRAKEN_WITHDRAW_FEE_XTZ
    else:
        raise Exception(f"{cryptocurrency} is not implemented")
    return Decimal(fee)


def get_liquidity_address(crypto: str) -> str:
    if crypto in [ETH, USDT, USDC]:
        return settings.KRAKEN_WITHDRAW_WALLET_ETH
    elif crypto == XRP:
        return settings.XRP_EXCHANGE_WALLET
    elif crypto in [BTC, BCH, LTC, DASH]:
        return ''
    elif crypto == XTZ:
        return settings.XTZ_EXCHANGE_WALLET


def get_liquidity_balance(address: str, crypto: str) -> Decimal:
    if address == 'liquidity':
        address = get_liquidity_address(crypto)
    if crypto == ETH:
        return get_eth_wallet_balance(address)
    elif crypto == XRP:
        return get_xrp_wallet_balance(address)
    elif crypto == BTC:
        return get_entire_btc_wallet_balance()
    elif crypto == LTC:
        return get_entire_ltc_wallet_balance()
    elif crypto == DASH:
        return get_entire_dash_wallet_balance()
    elif crypto == USDT:
        return get_usdt_wallet_balance(address)
    elif crypto == USDC:
        return get_usdc_wallet_balance(address)
    elif crypto == BCH:
        return get_entire_bch_wallet_balance()
    elif crypto == XTZ:
        return get_tezos_wallet_balance(address)


def get_received_payment(address: str, crypto: str) -> (Decimal, str):
    # TODO retrun txhash instead of None
    if crypto == ETH:
        return get_eth_wallet_balance(address), None
    elif crypto == XRP:
        return get_xrp_amount_and_hash(address)
    elif crypto == BTC:
        return get_btc_received_by_payment_list(address)
    elif crypto == LTC:
        return get_ltc_received_by_payment_list(address)
    elif crypto == DASH:
        return get_dash_received_by_payment_list(address)
    elif crypto == USDT:
        return get_usdt_wallet_balance(address), None
    elif crypto == USDC:
        return get_usdc_wallet_balance(address), None
    elif crypto == BCH:
        return get_bch_received_by_payment_list(address)
    elif crypto == XTZ:
        return get_tezos_wallet_balance(address), None


def get_liquidity_dict():
    cryptos_list = [c[1].lower() for c in AVAILABLE_CRYPTOCURRENCIES]
    result_dict = {}
    for crypto in cryptos_list:
        try:
            result_dict[crypto] = get_liquidity_balance('liquidity', crypto)
        except Exception as e:
            logger.error(e, exc_info=True)
            # Just pass, we don't want to block other crytos
    return result_dict


def get_transactions(
    crypto_type: str, query: Dict[str, Any]
) -> Union[List[Dict[str, Union[str, int, List[Dict[str, Union[str, int]]]]]], None]:
    """
    Get transactions from explorers
    @param crypto_type: crypto-currency name
    @type crypto_type: str
    @param query: a query for filtering transactions
    @type query: Dict[str, Any]
    @return: return transaction list
    @rtype: Union[List[Dict[str, Union[str, int, List[Dict[str, Union[str, int]]]]]], None]
    """
    if crypto_type == ETHEREUM_BASED_CRYPTO:
        transactions = EthExplorerApi().get_transactions(query=query)
    elif crypto_type == BITCOIN_BASED_CRYPTO:
        transactions = BtxExplorerApi().get_transactions(query=query)
    elif crypto_type == Transfer.XRP:
        transactions = XrpExplorerApi().get_transactions(query=query)
    else:
        raise NotImplementedError

    return transactions
