import re
from decimal import Decimal
from typing import Union

from bitcoin.core import b2lx, lx

from . import btc_conn
from interfaces.bitcoin.constants import BTC
from interfaces.bitcoin.utils import (
    get_received_by_payment_list, calculate_unspend_utxo
)


def is_btc_wallet_address(address: str) -> bool:
    """
    Helper method for BTC address validation
    """
    # https://en.bitcoin.it/wiki/List_of_address_prefixes
    btc_address_re = '^(bc1|tb1|m|n|[1-3])[a-zA-HJ-NP-Z0-9]{25,39}$'

    if not re.search(btc_address_re, address):
        return False
    return True


def btc2satoshi(value: Union[Decimal, float]) -> int:
    """
    Function converts btc to satoshi(with rounding to 8 digits after point).
    """
    return int(round(float(value) * 1e8, 8))


def satoshi2btc(value: int) -> float:
    """
    Function converts satoshi to btc.
    """
    return value / 1e8


def get_btc_wallet_balance(address: str) -> Decimal:
    amount = calculate_unspend_utxo(btc_conn, address)
    return Decimal(satoshi2btc(amount))


def get_btc_received_by_payment_list(address: str) -> (Decimal, str):
    return get_received_by_payment_list(btc_conn, BTC, address)


def get_entire_btc_wallet_balance() -> Decimal:
    return Decimal(satoshi2btc(btc_conn.getbalance(minconf=3)))


def create_btc_wallet() -> str:
    return btc_conn.getnewaddress()


def send_btc(address: str, amount: float) -> str:
    return b2lx(btc_conn.sendtoaddress(address, btc2satoshi(amount)))
