from decimal import Decimal
from typing import Union

from bitcoincash.core import b2lx
from cashaddress import convert

import bch_conn
from interfaces.bitcoin_cash.constants import BCH
from interfaces.bitcoin_cash.utils import (
    get_received_by_payment_list, calculate_unspend_utxo,
    clean_bitcoin_cash_address
)


def is_bch_wallet_address(address: str) -> bool:
    """
    Helper method for BCH address validation
    It uses special package becouse BCH has several addresses types.
    """
    return convert.is_valid(address) or convert.is_valid('bitcoincash:' + address)


def bch2satoshi(value: Union[Decimal, float]) -> int:
    """
    Function converts bch to satoshi(with rounding to 8 digits after point).
    """
    return int(round(float(value) * 1e8, 8))


def satoshi2bch(value: int) -> float:
    """
    Function converts satoshi to bch.
    """
    return value / 1e8


def get_bch_wallet_balance(address: str) -> Decimal:
    address = clean_bitcoin_cash_address(address)
    amount = calculate_unspend_utxo(bch_conn, address)
    return Decimal(satoshi2bch(amount))


def get_bch_received_by_payment_list(address: str) -> (Decimal, str):
    return get_received_by_payment_list(bch_conn, BCH, address)


def get_entire_bch_wallet_balance() -> Decimal:
    return Decimal(satoshi2bch(bch_conn.getbalance(minconf=3)))


def create_bch_wallet() -> str:
    return clean_bitcoin_cash_address(bch_conn.getnewaddress())


def send_bch(address: str, amount: float) -> str:
    return b2lx(bch_conn.sendtoaddress(address, bch2satoshi(amount)))
