import logging
import os
from time import sleep

from kuailian.bulk_tools.helpers import request_eth_from_wallet, ERC20_HELPER_OPTION
from interfaces.erc20.ethereum import web3_kraken
from interfaces.erc20.ethereum.infura import web3_infura
from eth_utils import ValidationError

logger = logging.getLogger(__name__)

module_dir = os.path.dirname(__file__)
with open(os.path.join(module_dir, 'erc20-abi.json')) as abif:
    erc20_abi = abif.read()

# TODO change those values to ERC-20 standard
REQUIRED_ERC20_GAS_TRANSFER = 58100
REQUIRED_ERC20_GAS_TRANSFER_NEW_ACCOUNT = 100000


class ERC20Contract:
    def __init__(self, contract_address: str,
                 required_gas_transfer: int = REQUIRED_ERC20_GAS_TRANSFER,
                 required_gas_transfer_new_account: int = REQUIRED_ERC20_GAS_TRANSFER_NEW_ACCOUNT,
                 contract_name='ERC20'
                 ):
        self._contract = web3_infura.eth.contract(
            address=web3_infura.toChecksumAddress(contract_address),
            abi=erc20_abi
        )
        self.required_gas_transfer = required_gas_transfer
        self.required_gas_transfer_new_account = required_gas_transfer_new_account
        self.contract_name = contract_name

    def get_balance(self, address: str):
        return self._contract.functions.balanceOf(address).call()

    def _get_required_gas_amount(self, receiver):
        erc20_balance = self.get_balance(receiver)
        return self.required_gas_transfer if erc20_balance > 0 \
            else self.required_gas_transfer_new_account

    def _request_value_from_tank_wallet(self, receiver, required_eth_balance):
        eth_balance = web3_kraken.eth.getBalance(receiver)
        request_value = required_eth_balance - eth_balance
        if request_value > 0:
            request_eth_from_wallet(receiver, request_value,
                                    from_wallet=ERC20_HELPER_OPTION,
                                    requester=self.contract_name)
            counter = 0
            # Additionally wait for balance
            while True:
                new_balance = web3_kraken.eth.getBalance(receiver)
                if new_balance >= required_eth_balance or counter > 30:
                    break
                counter += 1
                sleep(15)

    def transfer(self, wallet_from, _to: str, amount: int,
                 gas_price: int = None, nonce: int = None,
                 private_key=None):

        address = wallet_from if isinstance(wallet_from, str) else wallet_from.address
        if self.get_balance(address) < amount:
            raise ValidationError(f"""{self.contract_name}:Not enough tokens to send.""")

        if private_key is None:
            private_key = wallet_from.private_key
        if gas_price is None:
            gas_price: int = web3_kraken.eth.generateGasPrice()

        gas = self._get_required_gas_amount(_to)
        required_eth_balance = gas * gas_price
        # If needed
        self._request_value_from_tank_wallet(address, required_eth_balance)

        if nonce is None:
            nonce = web3_infura.eth.getTransactionCount(address)
        raw_tx = self._contract.functions.transfer(_to, amount) \
            .buildTransaction(
            {
                'nonce': nonce,
                'gasPrice': gas_price,
                'gas': gas
            }
        )
        signed_tx = web3_kraken.eth.account.signTransaction(raw_tx, private_key)
        # send via geth node because infura node could take more time to update wallet balance.
        tx_hex = web3_kraken.eth.sendRawTransaction(signed_tx.rawTransaction)
        return web3_kraken.toHex(tx_hex)
