import requests

from typing import Optional, Union, Dict, Any, List
from urllib.parse import urlencode

from django.conf import settings

from services_connector.core.handlers import response_handler, network_error_handler

from interfaces.ethereum.constants import ETH_NETWORKS


class EtherscanAPI:
    MAINNET = 'mainnet'
    API_TIMEOUT = 5

    def __init__(self, payload: Dict[str, Any], add_api_key: bool = True) -> None:
        """
        etherscan requester

        :param payload: GET params in dict format
        :type payload: dict
        :param add_api_key: adding api key to request
        :type add_api_key: bool
        """
        if add_api_key:
            payload['apikey'] = settings.ETHERSCAN_API_KEY
        self.payload = payload
        self.base_url = f'https://{self.api}.etherscan.io/api'

    @property
    def api(self) -> str:
        """
        return api url depends on env variable

        :return: api string
        :rtype: str
        """
        network = ETH_NETWORKS[self.MAINNET]
        if settings.INFURA_NETWORK in ETH_NETWORKS:
            network = ETH_NETWORKS[settings.INFURA_NETWORK]
        return network.api

    def _get_headers(self) -> Dict[str, str]:
        """
        return header depends on env variable
        Reason: CloudFlare firewall on rinkeby reject requests library user agent

        :return: header
        :rtype: dict
        """
        return {"user-agent": "curl"} if settings.INFURA_NETWORK != self.MAINNET else {}

    @network_error_handler(timeout=settings.ETHERSCAN_RETRY_TIMEOUT_SECONDS)
    def fetch(self) -> Union[Dict[str, Any], None]:
        """
        Fetching response from etherscan

        :return: None or response_data
        :rtype: Optional[dict]
        """
        params = urlencode(self.payload)
        url = f"{self.base_url}?{params}"

        response = requests.get(url=url, headers=self._get_headers(), timeout=self.API_TIMEOUT)
        msg = f"etherscan: {params}"
        is_error, data = response_handler(response=response, msg_text=msg)

        if is_error:
            return None
        else:
            return data


def get_all_transactions_for_address(
    address: str, start_block: int = 0, end_block: int = 99999999
) -> Union[list, bool]:
    normal_payload = dict(
        module='account',
        action='txlist',
        address=address.lower(),
        startblock=start_block,
        endblock=end_block,
        sort='asc',
    )
    normal_txs = EtherscanAPI(normal_payload).fetch()
    if normal_txs is None:
        return False
    normal_txs = normal_txs.get('result', [])
    if not isinstance(normal_txs, list):
        return False
    internal_payload = dict(
        module='account',
        action='txlistinternal',
        address=address.lower(),
        startblock=start_block,
        endblock=end_block,
        sort='asc',
    )
    internal_txs = EtherscanAPI(internal_payload).fetch()
    if internal_txs is None:
        return False
    internal_txs = internal_txs.get('result', [])
    if not isinstance(internal_txs, list):
        return False
    return normal_txs + internal_txs


def get_etherscan_transactions_by_wallets(
    from_wallet: str, target_wallets: List[str]
) -> Optional[list]:
    """

    @param from_wallet:
    @type from_wallet: str
    @param target_wallets:
    @type target_wallets: List[str]
    @return:
    @rtype:
    """
    if not isinstance(target_wallets, list):
        raise TypeError

    payload = dict(
        module='account',
        action='txlist',
        address=from_wallet,
        startblock=0,
        endblock=99999999,
        sort='asc',
    )
    res_dict = EtherscanAPI(payload).fetch()

    transactions = []
    if res_dict and res_dict['status'] == '1':
        target_wallets = [wallet.lower() for wallet in target_wallets]
        for res in res_dict['result']:
            if res['from'].lower() == from_wallet.lower() and res['to'].lower() in target_wallets:
                transactions.append(res)
        return transactions

    else:
        return
