import logging
import requests
from decimal import Decimal
from typing import Dict, List, Union, Any
from urllib.parse import urlencode, urljoin

from django.conf import settings
from eth_utils.currency import from_wei
from services_connector.core.handlers import (
    network_error_handler, response_handler
)
from interfaces.constants import ETH, USDC, USDT
from interfaces.tether.helpers import microcents2usdt
from interfaces.usdc.helpers import mircousd2usdc

__all__ = ['EthExplorerApi', ]
logger = logging.getLogger(__name__)


class EthExplorerApi:
    @network_error_handler(timeout=2)
    def get_transactions(
        self, query: Dict[str, Any]
    ) -> Union[List[Dict[str, Union[str, int, List[Dict[str, Union[str, int]]]]]], None]:
        """
        Get transactions which was made from wallet to internal service wallets.

        :param query:
        :type query: str
        :return: transactions which was made from invoice wallet
        :rtype: Union[List[Dict[str, Any], None]
        """

        query_params = urlencode(query)
        msg = f"get transaction by {query_params} from eth-explorer"
        logger.info(f"Start {msg}")
        url = urljoin(
            settings.ETH_EXPLORER_API_URL,
            urljoin(
                "api/transactions/",
                f"?{query_params}"
            )
        )

        response = requests.get(url=url, timeout=settings.EXPLORERS_API_TIMEOUT)
        is_error, data = response_handler(response=response, msg_text=msg)

        if is_error:
            return None
        else:
            return data


class EthTransactionAdapter:
    def get_converted_value(self, crypto: str, value: int) -> Union[float, Decimal]:
        if crypto == ETH:
            return from_wei(number=value, unit='ether')
        elif crypto == USDC:
            return mircousd2usdc(value=value)
        elif crypto == USDT:
            return microcents2usdt(value=value)
        else:
            raise NotImplementedError

    def adapt(self, transactions, crypto: str):
        for trx in transactions:
            new_value = self.get_converted_value(crypto=crypto, value=trx['value'])
            trx['value'] = Decimal(new_value)

        return transactions
