import time
from decimal import Decimal

from django.conf import settings
from django.utils.crypto import get_random_string
from web3.exceptions import InvalidAddress

import web3_kraken
from invoices.models import Wallet
from interfaces.ethereum.constants import ETH_NETWORKS, CONTRACT_ABI
from interfaces.ethereum.geth import web3_geth
from interfaces.ethereum.infura import web3_infura


def send_eth(
    wallet_from: Wallet,
    _to: str,
    wei_amount: int,
    send_value_without_fee: bool = False,
    nonce: int = None,
    gas_price: int = None,
    private_key='',
    retry: int = 0
) -> (str, int):
    """ send ether from temporary wallet to the some company wallet """
    if isinstance(wallet_from, str):
        address = wallet_from
    else:
        address = wallet_from.address
    if not private_key:
        private_key = wallet_from.private_key

    if gas_price is None:
        gas_price: int = web3_geth.eth.generateGasPrice()

    if send_value_without_fee:
        send_value = wei_amount
    else:
        send_value = wei_amount - gas_price * 21000

    if nonce is None:
        nonce = web3_infura.eth.getTransactionCount(address)

    if send_value > 0:
        signed_txn = web3_geth.eth.account.signTransaction(dict(
            nonce=nonce,
            gasPrice=gas_price,
            gas=21000,
            to=_to,
            value=send_value,
            data=b'',
        ),
            private_key,
        )

        try:
            tx_hex = web3_infura.eth.sendRawTransaction(
                signed_txn.rawTransaction
            )
        except ValueError as err:
            """
            it seems problem with gas price. It was changed
            https://stackoverflow.com/questions/46349202/rinkeby-replacement-transaction-underpriced

            in case with automation process can be situation
            when we try send money before get them from Kraken
            """
            error_msgs = [
                'transaction underpriced',
                'insufficient funds for gas * price + value'
            ]
            if (err.args[0]['message'] in error_msgs and retry < 5):
                # transaction can be pending
                time.sleep(10)
                return send_eth(
                    wallet_from,
                    _to,
                    wei_amount,
                    send_value_without_fee,
                    nonce,
                    private_key=private_key,
                    retry=retry + 1
                )
            else:
                raise ValueError(err)

        res = web3_geth.toHex(tx_hex)
        # TODO: we need subtract sending amount
        if isinstance(wallet_from, Wallet):
            wallet_from.wei_balance = 0
            wallet_from.save(update_fields=['wei_balance'])
    else:
        res = f'The balance is too low to send it. ' \
              f'The wallet {address}, with {send_value} wei'
    return res, gas_price


def send_eth_from_withdraw_wallet(_to: str,
                                  wei_amount: int,
                                  send_value_without_fee: bool = False,
                                  gas_price: int = None):
    if gas_price is None:
        gas_price: int = web3_geth.eth.generateGasPrice()

    if send_value_without_fee:
        send_value = wei_amount
    else:
        send_value = wei_amount - gas_price * 21000

    web3_kraken.geth.personal.unlockAccount(
        settings.KRAKEN_WITHDRAW_WALLET_ETH,
        settings.KRAKEN_WITHDRAW_WALLET_ETH_PASSPHRASE
    )
    tx_hash = web3_kraken.eth.sendTransaction({
        'to': _to,
        'from': settings.KRAKEN_WITHDRAW_WALLET_ETH,
        'value': send_value,
        'gasPrice': gas_price
    })
    return web3_geth.toHex(tx_hash)


def get_network_api() -> str:
    """
    returns network id based on geth network version
    """
    return ETH_NETWORKS[settings.GETH_NETWORK].api


def get_network_url() -> str:
    """
    returns network id based on geth network version
    """
    return ETH_NETWORKS[settings.GETH_NETWORK].url


def get_contract_address() -> str:
    """
    returns smart contract address based on geth network version
    """
    return ETH_NETWORKS[settings.GETH_NETWORK].contract


def get_contract_instance():
    """ init ethereum bulk tool smart contract instance """
    contract_address = get_contract_address()
    if contract_address:
        contract = web3_geth.eth.contract(
            address=contract_address, abi=CONTRACT_ABI
        )
        return contract
    return False


def create_eth_wallet(fake: bool = False) -> Wallet:
    """ creates intermediate wallet to allow payment validation """
    password = get_random_string(length=20)
    wallet = web3_infura.eth.account.create(password)
    wallet_obj = Wallet.objects.create(
        address=wallet.address,
        password=password,
        private_key=web3_infura.toHex(wallet.privateKey),
        fake=fake
    )
    return wallet_obj


def is_eth_wallet_address(address: str) -> bool:
    """
    validate that is given address is account
    """
    return web3_infura.isAddress(address.lower())


def is_eth_contract(address: str) -> bool:
    """
    validate that is given address is contract address
    """
    try:
        return len(web3_infura.eth.getCode(web3_infura.toChecksumAddress(address))) > 0
    except InvalidAddress:
        return False


def get_eth_wallet_balance(address: str) -> Decimal:
    wei_balance = web3_infura.eth.getBalance(address)
    balance = web3_infura.fromWei(wei_balance, 'ether')
    return balance
