from urllib3.exceptions import HTTPError

from requests.exceptions import RequestException
from twilio.rest import Client
from twilio.base.exceptions import TwilioException
from eth_utils.toolz import assoc
from constance import config

from django.conf import settings
from django.core.cache import cache

from interfaces.ethereum.strategies import set_strategy


class FailedRequestNotificationMiddleware:
    """
    Send sms notification when request to node is failed
    """
    message = None
    cache_key = 'web3_node_failed_notification'

    def __init__(self, make_request, w3):
        self.w3 = w3
        self.make_request = make_request

    def __call__(self, method, params):
        try:
            response = self.make_request(method, params)
            return response
        except RequestException as err:

            if not cache.get(self.cache_key) and self.message is not None:
                try:
                    client = Client(
                        settings.TWILIO_ACCOUNT_SID, settings.TWILIO_AUTH_TOKEN
                    )
                    for number in settings.WEB3_SMS_NOTIFICATION_NUMBERS:
                        client.messages.create(
                            to=number,
                            from_=settings.TWILIO_CALLER_ID,
                            body=self.message
                        )
                except TwilioException:
                    pass

                cache.set(self.cache_key, 'sent', 60 * 15)

            raise err


class PlatformFailedRequestNotificationMiddleware(FailedRequestNotificationMiddleware):
    """
    Send sms notifications when platform request to node is failed
    """
    message = 'Something wrong with platform node'
    cache_key = 'web3_node_failed_notification_platform'


class BankFailedRequestNotificationMiddleware(FailedRequestNotificationMiddleware):
    """
    Send sms notifications when bank request to node is failed
    """
    message = 'Something wrong with bank node'
    cache_key = 'web3_node_failed_notification_bank'


class RetrySendRequestMiddleware:
    """
    Retry sent request to slave node
    """
    slave_node_url = None

    def __init__(self, make_request, w3):
        self.w3 = w3
        self.make_request = make_request

    def __call__(self, method, params):
        try:
            response = self.make_request(method, params)
            return response
        except (RequestException, HTTPError):
            return self.retry(method, params)
        except ValueError as err:
            if err.args[0]['message'] == 'no suitable peers available':
                return self.retry(method, params)
            raise err

    def retry(self, method, params):
        self.w3.provider.endpoint_uri = self.slave_node_url
        return self.make_request(method, params)


class PlatformRetrySendRequestMiddleware(RetrySendRequestMiddleware):
    """
    Retry sent request to platform slave node
    """
    slave_node_url = settings.GETH_SLAVE_NODE_URL


class BankRetrySendRequestMiddleware(RetrySendRequestMiddleware):
    """
    Retry sent request to bank slave node
    """
    slave_node_url = settings.GETH_SLAVE_KRAKEN_NODE_URL


class GasPriceStrategyMiddleware:
    """
    Replace default gas price strategy middleware with custom.
    Gas price strategy should be set from config
    """

    def __init__(self, make_request, w3):
        self.w3 = w3
        self.make_request = make_request

    def __call__(self, method, params):
        if method == 'eth_sendTransaction':
            transaction = params[0]
            if 'gasPrice' not in transaction:
                self.set_gas_price_strategy()
                generated_gas_price = self.w3.eth.generateGasPrice(transaction)
                if generated_gas_price is not None:
                    transaction = assoc(
                        transaction, 'gasPrice', generated_gas_price
                    )
                    return self.make_request(method, [transaction])
        return self.make_request(method, params)

    def set_gas_price_strategy(self):
        raise NotImplemented


class DailyPaymentsStrategyMiddleware(GasPriceStrategyMiddleware):
    """
    Set strategy for daily payments (web3_geth)
    """

    def set_gas_price_strategy(self):
        self.w3 = set_strategy(self.w3, config.GAS_STRATEGY_DAILY_PAYMENTS)


class InvoicesStrategyMiddleware(GasPriceStrategyMiddleware):
    """
    Set strategy for invoices (web3_infura)
    """

    def set_gas_price_strategy(self):
        self.w3 = set_strategy(self.w3, config.GAS_STRATEGY_INVOICES)


class BankStrategyMiddleware(GasPriceStrategyMiddleware):
    """
    Set strategy for bank (web3_kraken)
    """

    def set_gas_price_strategy(self):
        self.w3 = set_strategy(self.w3, config.GAS_STRATEGY_BANK)
