import logging
from decimal import Decimal
from typing import Tuple

import base58
from django.conf import settings

from bank.models import DestinationPayment, Transfer
from . import AccountRepeat as Account, RippleDataAPIClientRepeat as RippleDataAPIClient

logger = logging.getLogger(__name__)

DEFAULT_RIPPLE_FEE = Decimal('0.00001')
RIPPLE_ALPHABET = b'rpshnaf39wBUDNEGHJKLM4PQRST7VWXYZ2bcdeCg65jkm8oFqi1tuvAxyz'


def send_xrp(
    account_id: str,
    seed: str,
    destination: str,
    amount: Decimal,
    destination_tag: str = '',
    pay_with_fee=False
) -> str or None:
    """
    Send XRP
    """
    pay_amount = amount - DEFAULT_RIPPLE_FEE if pay_with_fee else amount
    account = Account(settings.RIPPLE_NETWORK_URL, account_id, seed)
    tx_json = {
        "TransactionType": "Payment",
        "Account": account_id,
        "Destination": destination,
        "Amount": str(round(pay_amount * account.xrp_base, 0)),
    }
    if destination_tag:
        tx_json.update({"DestinationTag": destination_tag})

    logger.info(f"Sending bought crypto to users wallet, details:{tx_json}")
    response = account.sign_and_submit(
        tx_json=tx_json,
        secret=seed,
    )
    logger.info(f"Send crypto to user response:{response}")

    if response['status'] == 'success':
        return response['tx_json']['hash']
    # We need more info if something goes wrong
    raise ValueError(str(response))


def is_xrp_wallet_address(address: str) -> bool:
    """
    Helper method for XRP address validation
    """
    base58.alphabet = RIPPLE_ALPHABET
    try:
        base58.b58decode_check(address)
        return True
    except ValueError:
        return False


def get_xrp_amount_and_hash(destination_tag: str) -> Tuple[Decimal, str]:
    account_id = settings.XRP_EXCHANGE_WALLET
    data_client = RippleDataAPIClient()
    params = {
        'destination_tag': destination_tag
    }
    payment_history = data_client.get_account_payments(
        address=account_id, **params
    )
    transfer = Transfer.objects.filter(sender_tag=destination_tag).first()
    for payment in payment_history['payments']:
        dest_payment, _ = DestinationPayment.objects.get_or_create(
            tx_hash=payment['tx_hash'],
            cryptocurrency=DestinationPayment.XRP,
            defaults={'transfer': transfer, 'amount': Decimal(payment['amount'])}
        )
        if dest_payment.is_processed:
            continue
        return Decimal(payment['amount']), payment['tx_hash']

    return Decimal(0.0), ''


def get_xrp_wallet_balance(address: str) -> Decimal:
    rp = RippleDataAPIClient()
    balance = rp.get_account_balances(
        address, currency='XRP')
    if 'result' in balance:
        if balance['result'] == 'success':
            return Decimal(balance['balances'][0]['value'])
    return Decimal(0.0)
