from decimal import Decimal
from typing import Union

from pytezos import RpcError
from pytezos.crypto import Key
from pytezos.encoding import scrub_input, base58_decode
from pytezos.operation.content import ContentMixin

from interfaces.tezos import tezos_conn, tezos_client_spawn
from invoices.models import Wallet

DEFAULT_FEE = 1285


def is_tezos_wallet_address(address: str) -> bool:
    """
    Helper method for Tezos address validation
    """
    addr_len = len(address)
    if addr_len not in (36, 35):
        return False
    if not any(address.startswith(prefix) for prefix in ('tz1', 'tz2', 'tz3', 'KT1')):
        return False
    try:
        base58_decode(scrub_input(address))
        return True
    except: # TODO - do not use bare 'except'
        return False


def tez2mutez(value: Union[Decimal, float]) -> int:
    """
    Function converts tezos to microcents(with rounding to 8 digits after point).
    """
    return int(round(float(value) * 1e6, 6))


def mutez2tez(value: int) -> float:
    """
    Function converts microcents to tezos.
    """
    return value / 1e6


def get_tezos_wallet_balance(address: str) -> Decimal:
    balance = tezos_conn.account(address)['balance']
    balance = int(balance)
    return Decimal(mutez2tez(balance))


def create_tezos_wallet(fake: bool = False) -> Wallet:
    key = Key.generate(export=False)
    wallet_obj = Wallet.objects.create(
        address=key.public_key_hash(),
        password='',
        private_key=key.secret_key(),
        fake=fake,
        cryptocurrency=Wallet.XTZ
    )
    return wallet_obj


def is_key_revealed(key):
    client = tezos_client_spawn(key)
    try:
        client.reveal().autofill()
        return False
    except RpcError as e:
        if 'previously_revealed_key' in str(e):
            return True
        else:
            raise e


def make_transaction_operation(sender_key, destination, amount, is_revealed=True, send_value_without_fee=True):
    client = tezos_client_spawn(sender_key)
    content = ContentMixin()
    if not send_value_without_fee:
        amount = amount - DEFAULT_FEE - DEFAULT_FEE * is_revealed
    operation_dict = content.transaction(destination, amount, source=sender_key.public_key_hash())
    if is_revealed is False:
        reveal_dict = {
            'kind': 'reveal',
            'source': sender_key.public_key_hash(),
            'fee': '0',
            'counter': '0',
            'gas_limit': '0',
            'storage_limit': '0',
            'public_key': sender_key.public_key()
        }
        operation = client.operation_group(contents=[reveal_dict, operation_dict])
    else:
        operation = client.operation(operation_dict)

    return operation


def decode_secret_key(sk_hex):
    secret_key = base58_decode(scrub_input(sk_hex)).hex()
    return Key.from_secret_exponent(bytes.fromhex(secret_key))


def send_tezos(wallet_from, to_address: str, amount: float, send_value_without_fee=True) -> str:
    sk_hex = wallet_from if isinstance(wallet_from, str) else wallet_from.private_key
    sender_key = decode_secret_key(sk_hex)
    send_amount = tez2mutez(amount)
    operation = make_transaction_operation(sender_key, to_address, send_amount,
                                           is_revealed=is_key_revealed(sender_key),
                                           send_value_without_fee=send_value_without_fee)

    return operation.fill().sign().inject()['hash']
