from decimal import Decimal
from typing import Dict, Any

from django.conf import settings

from bank.models import Transfer, DestinationPayment
from interfaces.explorer import BtxExplorerApi
from interfaces.constants import BITCOIN_BASED_CRYPTO


def clean_bitcoin_cash_address(address):
    address = str(address)
    if address.startswith('bitcoincash:'):
        address = address[12:]
    return address


def get_address_income(address, txs):
    for tx in txs:
        if clean_bitcoin_cash_address(tx['address']) == \
            clean_bitcoin_cash_address(address) and tx['category'] == 'receive':
            return tx['amount']

    raise ValueError("No income transaction found")


def get_received_by_payment_list(conn, cryptocurrency, address: str or bytes) -> (Decimal, str):
    transfer = Transfer.objects.filter(wallet_address=address).first()

    address = clean_bitcoin_cash_address(address)
    unspend = conn.listreceivedbyaddress(address_filter=address, minconf=3)
    if len(unspend) > 0:
        txs = unspend[0]['txids']
        for txid in txs:
            transaction_overview = conn.gettransaction(txid)
            income_amount = get_address_income(address, transaction_overview['details'])
            dest_payment, _ = DestinationPayment.objects.get_or_create(
                tx_hash=txid,
                cryptocurrency=cryptocurrency,
                defaults={'transfer': transfer, 'amount': Decimal(income_amount)}
            )
            if dest_payment.is_processed:
                continue
            return Decimal(income_amount), txid

    return Decimal(0.0), None


def calculate_unspend_utxo(conn, address: str):
    unspend = conn.listunspent(addrs=[address], minconf=3)
    amount = 0
    for utx in unspend:
        amount += utx['amount']
    return amount


def get_query_with_sender_n_receiver_address_for_getting_transactions(
    transfer: Transfer, old_kraken_eth_address: bool = True
) -> Dict[str, Any]:
    """
    Prepare a query for getting transactions list from explorer.
    Transactions will be filtered by sender address and receiver address
    @param transfer: Transfer object
    @type transfer: Transfer
    @param old_kraken_eth_address: Use kraken_eth_address or
    @type old_kraken_eth_address: bool
    @return: Query like a dictionary
    @rtype: Dict[str, Any]
    """
    if transfer.cryptocurrency == Transfer.ETH:
        if old_kraken_eth_address:
            address_from = settings.KRAKEN_WITHDRAW_WALLET_ETH_OLD
        else:
            address_from = settings.KRAKEN_WITHDRAW_WALLET_ETH
        query = {
            "address_from": address_from,
            "address_to": transfer.wallet_address,
        }
    elif transfer.cryptocurrency in BITCOIN_BASED_CRYPTO:
        if transfer.cryptocurrency == Transfer.BTC:
            sender_address = settings.BTC_EXCHANGE_WALLET
        elif transfer.cryptocurrency == Transfer.LTC:
            sender_address = settings.LTC_EXCHANGE_WALLET
        elif transfer.cryptocurrency == Transfer.DASH:
            sender_address = settings.DASH_EXCHANGE_WALLET
        elif transfer.cryptocurrency == Transfer.BCH:
            sender_address = settings.BCH_EXCHANGE_WALLET
        else:
            raise NotImplementedError

        block_type = BtxExplorerApi.BTX_CURRENCY_MAPPING[transfer.cryptocurrency]
        query = {
            "inputs__address": transfer.wallet_address,
            "outputs__address": sender_address,
            "transaction__block__type": block_type
        }
    elif transfer.cryptocurrency == Transfer.XRP:
        query = {
            "sender": settings.XRP_EXCHANGE_WALLET,
            "receiver": transfer.wallet_address,
        }
    else:
        raise NotImplementedError
    return query
