from django.db import models
from django_extensions.db.models import TimeStampedModel
from encrypted_model_fields.fields import EncryptedCharField

from bank.models import MarkupFee

from invoices.managers import EthExchangeRateManager


class Wallet(models.Model):
    """
    Crypto wallet model
    """
    ETH = 'eth'
    USDC = 'usdc'
    USDT = 'usdt'
    XTZ = 'xtz'

    CRYPTOCURRENCY_CHOICES = (
        (ETH, 'Ethereum'),
        (USDC, 'USDC'),
        (USDT, 'USDT'),
        (XTZ, 'Tezos'),
    )

    address = models.CharField('Address', max_length=42)
    private_key = EncryptedCharField('Private key', max_length=70)
    password = EncryptedCharField('Password', max_length=20)
    fake = models.BooleanField('Fake', default=False)

    cryptocurrency = models.CharField(
        'Cryptocurrency',
        max_length=10,
        default=ETH,
        choices=CRYPTOCURRENCY_CHOICES
    )
    # TODO: rename to balannce
    wei_balance = models.DecimalField(
        'Wei balance',
        default=0,
        max_digits=25,
        decimal_places=0
    )

    class Meta:
        verbose_name = 'wallet'
        verbose_name_plural = 'Wallets'

    def __str__(self):
        return self.address

    @staticmethod
    def get_wallet_for_address(address, raise_exception=False):
        wallet = Wallet.objects.filter(address=address).first()
        if raise_exception and not wallet:
            raise Wallet.DoesNotExist(f"Wallet for address: {address} doesn't exist.")
        return wallet


class EthExchangeRate(TimeStampedModel):
    """
    Stores exchange rate for ETH/USD, EHT/BTC
    """
    EUR = 'eur'
    USD = 'usd'
    BTC = 'btc'

    # SELLING
    ETH_EUR = 'ETH/EUR'
    ETH_USD = 'ETH/USD'
    ETH_BTC = 'ETH/BTC'
    ETH_USDC = 'ETH/USDC'
    ETH_USDT = 'ETH/USDT'
    BTC_EUR = 'BTC/EUR'
    BTC_USD = 'BTC/USD'
    LTC_EUR = 'LTC/EUR'
    LTC_USD = 'LTC/USD'
    XRP_EUR = 'XRP/EUR'
    XRP_USD = 'XRP/USD'
    DASH_EUR = 'DASH/EUR'
    DASH_USD = 'DASH/USD'
    USDT_EUR = 'USDT/EUR'
    USDT_USD = 'USDT/USD'
    BCH_EUR = 'BCH/EUR'
    BCH_USD = 'BCH/USD'
    USDC_EUR = 'USDC/EUR'
    USDC_USD = 'USDC/USD'
    XTZ_EUR = 'XTZ/EUR'
    XTZ_USD = 'XTZ/USD'

    # BUYING
    EUR_ETH = 'EUR/ETH'
    EUR_BTC = 'EUR/BTC'
    USD_ETH = 'USD/ETH'
    USD_BTC = 'USD/BTC'
    BTC_ETH = 'BTC/ETH'
    EUR_LTC = 'EUR/LTC'
    USD_LTC = 'USD/LTC'
    EUR_XRP = 'EUR/XRP'
    USD_XRP = 'USD/XRP'
    EUR_DASH = 'EUR/DASH'
    USD_DASH = 'USD/DASH'
    EUR_USDT = 'EUR/USDT'
    USD_USDT = 'USD/USDT'
    EUR_BCH = 'EUR/BCH'
    USD_BCH = 'USD/BCH'
    EUR_USDC = 'EUR/USDC'
    USD_USDC = 'USD/USDC'
    EUR_XTZ = 'EUR/XTZ'
    USD_XTZ = 'USD/XTZ'
    USDC_ETH = 'USDC/ETH'
    USDT_ETH = 'USDT/ETH'

    TYPES_CHOICES = (
        (USD, 'USD'),
        (BTC, 'BTC'),
        (EUR, 'EUR'),
        (ETH_EUR, 'ETH/EUR'),
        (ETH_USD, 'ETH/USD'),
        (ETH_BTC, 'ETH/BTC'),
        (BTC_EUR, 'BTC/EUR'),
        (BTC_USD, 'BTC/USD'),
        (EUR_ETH, 'EUR/ETH'),
        (EUR_BTC, 'EUR/BTC'),
        (EUR_XRP, 'EUR/XRP'),
        (USD_ETH, 'USD/ETH'),
        (USD_BTC, 'USD/BTC'),
        (BTC_ETH, 'BTC/ETH'),
        # LTC
        (LTC_EUR, 'LTC/EUR'),
        (LTC_USD, 'LTC/USD'),
        (EUR_LTC, 'EUR/LTC'),
        (USD_LTC, 'USD/LTC'),
        # XRP
        (XRP_EUR, 'XRP/EUR'),
        (XRP_USD, 'XRP/USD'),
        (USD_XRP, 'USD/XRP'),
        (EUR_XRP, 'EUR/XRP'),
        # DASH
        (DASH_EUR, 'DASH/EUR'),
        (DASH_USD, 'DASH/USD'),
        (EUR_DASH, 'EUR/DASH'),
        (USD_DASH, 'USD/DASH'),
        # Tether
        (USDT_EUR, 'USDT/EUR'),
        (USDT_USD, 'USDT/USD'),
        (EUR_USDT, 'EUR/USDT'),
        (USD_USDT, 'USD/USDT'),
        (USDT_ETH, 'USDT/ETH'),
        (ETH_USDT, 'ETH/USDT'),
        # Bitcon Cash
        (BCH_EUR, 'BCH/EUR'),
        (BCH_USD, 'BCH/USD'),
        (EUR_BCH, 'EUR/BCH'),
        (USD_BCH, 'USD/BCH'),
        # USD Coin
        (USDC_EUR, 'USDC/EUR'),
        (USDC_USD, 'USDC/USD'),
        (EUR_USDC, 'EUR/USDC'),
        (USD_USDC, 'USD/USDC'),
        (USDC_ETH, 'USDC/ETH'),
        (ETH_USDC, 'ETH/USDC'),
        # Tezos
        (XTZ_EUR, 'XTZ/EUR'),
        (XTZ_USD, 'XTZ/USD'),
        (EUR_XTZ, 'EUR/XTZ'),
        (USD_XTZ, 'USD/XTZ'),
    )

    COINGECKO = 'coingecko'
    KRAKEN = 'kraken'

    PROVIDER_CHOICES = (
        (COINGECKO, 'Coingecko'),
        (KRAKEN, 'Kraken')

    )
    provider = models.CharField(
        'Provider',
        max_length=10,
        default=COINGECKO,
        choices=PROVIDER_CHOICES
    )
    rate = models.DecimalField('Rate', max_digits=23, decimal_places=18)
    exchange_type = models.CharField(
        'Type',
        max_length=10,
        default=USD,
        choices=TYPES_CHOICES
    )

    objects = EthExchangeRateManager()

    class Meta:
        db_table = 'Eth2UsdExchangeRate'
        verbose_name = 'rate'
        verbose_name_plural = 'EthExchange'

    def __str__(self):
        return f'{self.rate} - {self.exchange_type}: {self.created}'

    @classmethod
    def get_transaction_price_for_pair(cls, pair, transaction, old=False, with_obj=False):
        if transaction == 'sell':
            fee_type = MarkupFee.SELL_INVISIBLE_FEE
        else:
            fee_type = MarkupFee.BUY_INVISIBLE_FEE
        invisible_fee = MarkupFee.objects.get(type=fee_type).percent_value
        pair_rate = cls.objects.filter(exchange_type=pair).first()
        if old:
            pair_rate = cls.objects.filter(
                exchange_type=pair,
                created__lte=pair_rate.created - timedelta(hours=24)
            ).first()
            if not pair_rate:  # in case if we don't have 1 day old exchange rates
                pair_rate = cls.objects.filter(
                    exchange_type=pair,
                ).order_by('created').first()

        rate = pair_rate.rate + (pair_rate.rate / 100) * invisible_fee
        rate = Decimal(rate)
        if with_obj:
            return rate, pair_rate
        return rate

