from django.contrib.auth.models import UserManager
from django.db.models import Manager


class UserSoftDeletionManager(UserManager):
    def __init__(self, *args, **kwargs):
        self.alive_only = kwargs.pop('alive_only', True)
        super().__init__(*args, **kwargs)

    def get_queryset(self):
        if self.alive_only:
            return super().get_queryset().filter(deleted_at=None)
        return super().get_queryset()



class UserTreeManager(Manager):
    use_in_migrations = True
    cache_key = 'users_tree'

    def get_tree(self, ignore_cache=False):
        references = cache.get(self.cache_key, None)

        if references is None or ignore_cache:
            references = dict()
            roots = []
            for item in super().get_queryset().all():
                # get or make a new node
                if item.pk in references:
                    n = references[item.pk]
                    n.item = item
                else:
                    n = Node(children=[], item=item)
                    references[item.pk] = n

                if item.referral_id is None:
                    # if item is root (no parent)
                    roots.append(n)
                else:
                    # if item has a parent
                    if item.referral_id in references:
                        # item parent already seen
                        parent_n = references[item.referral_id]
                        parent_n.add_child(n)
                    else:
                        # item not seen yet
                        parent_n = Node(children=[n], item=None)
                        references[item.referral_id] = parent_n

            cache.set(self.cache_key, references, 60*60)

        return references

    def dfs(self, root, levels, affiliates_in_use, level=0):

        if level > 0 and root.item.id not in affiliates_in_use:
            levels[level].append(root.item)
            affiliates_in_use.append(root.item.id)

        for node in root.children:
            if level < settings.AFFILIATE_LEVELS:
                self.dfs(node, levels, affiliates_in_use, level + 1)

        return levels

    def bfs(self, node, levels, all_levels=False):
        result = []
        node.level = 0
        queue = deque([node])

        while queue:
            node = queue.popleft()
            result.append(node.item.id)
            if node.level > 0:
                levels[node.level].append(node.item)

            if all_levels or node.level < settings.AFFILIATE_LEVELS:
                for child in node.children:
                    child.level = node.level + 1
                    if child.item.id not in result and child not in queue:
                        queue.append(child)

        return levels

    def get_affiliates(self, user_id, tree=None):
        if tree is None:
            return self.get_affiliates_from_db(user_id)
        levels = {i: [] for i in range(1, settings.AFFILIATE_LEVELS + 1)}
        if user_id in tree:
            root = tree[user_id]
        else:
            return levels

        return self.bfs(root, levels)

    def get_all_affiliates(self, user_id, tree=None):
        if tree is None:
            tree = self.get_tree()
        levels = {i: [] for i in range(1, 32767)}   # small integer field
        if user_id in tree:
            root = tree[user_id]
        else:
            return levels

        return self.bfs(root, levels, True)

    @staticmethod
    def get_affiliate_licenses(affiliates):
        licenses_by_level = {str(i): 0 for i in range(1, settings.AFFILIATE_LEVELS + 1)}

        for lvl in affiliates:
            for affiliate in affiliates[lvl]:
                licenses = affiliate.total_lr_licenses + affiliate.total_mr_licenses + affiliate.total_hr_licenses
                licenses_by_level[str(lvl)] += licenses

        return licenses_by_level

    def get_referrals(self, user_id, tree=None):
        if tree is None:
            return self.get_referrals_from_db(user_id)
        levels = {i: None for i in range(1, settings.AFFILIATE_LEVELS + 1)}
        if user_id in tree:
            root = tree[user_id]
        else:
            return levels

        referrals_in_use = [user_id]
        if root.item.referral_id:
            for level in range(1, settings.AFFILIATE_LEVELS + 1):
                referral_id = root.item.referral_id
                if not referral_id:
                    break
                if referral_id in tree:
                    if referral_id not in referrals_in_use:
                        root = tree[referral_id]
                        levels[level] = root.item
                        referrals_in_use.append(referral_id)
                    else:
                        break
                else:
                    break

        return levels

    def update_user_referrals(self, user):
        if user.referral_id:
            tree = self.get_tree()
            if user.id in tree:
                tree[user.id].update_item(user)
                cache.set(self.cache_key, tree, 60 * 20)
            else:
                tree = self.get_tree(ignore_cache=True)

            referrals = self.get_referrals(user.id, tree)
            update_referrals = []

            for referral in referrals.values():
                if not referral:
                    break
                affiliates = self.get_affiliates(referral.id, tree)
                affiliate_licenses = self.get_affiliate_licenses(affiliates)
                del affiliate_licenses['10']
                referral.affiliate_licenses = sum(affiliate_licenses.values())
                update_referrals.append(referral)

            self.bulk_update(update_referrals, ['affiliate_licenses'])

    @staticmethod
    def affiliate_ids_cache_key(user_id):
        return f'affiliate_ids_{user_id}'

    @staticmethod
    def referral_ids_cache_key(user_id):
        return f'referral_ids_{user_id}'

    def get_affiliate_ids(self, user_id, flat=False, tree=None):
        affiliate_ids = cache.get(self.affiliate_ids_cache_key(user_id), None)
        if affiliate_ids is None:
            affiliate_ids = self.set_affiliate_ids_cache(user_id, tree)

        if flat:
            affiliate_ids = [item for sublist in affiliate_ids.values() for item in sublist]

        return affiliate_ids

    def set_affiliate_ids_cache(self, user_id, tree=None):
        affiliate_ids = {i: [] for i in range(1, settings.AFFILIATE_LEVELS + 1)}
        if tree:
            affiliates = self.get_affiliates(user_id, tree)
            for lvl in affiliates:
                affiliate_ids[lvl] = [affiliate.id for affiliate in affiliates[lvl] if affiliate.is_active]
        else:
            for user in self._affiliates_sql(user_id):
                if user.level > 0:
                    affiliate_ids[user.level].append(user.id)

        cache.set(self.affiliate_ids_cache_key(user_id), affiliate_ids, 60*60*3)
        return affiliate_ids

    def get_referral_ids(self, user_id, tree=None):
        referral_ids = cache.get(self.referral_ids_cache_key(user_id), None)
        if referral_ids is None:
            referral_ids = self.set_referral_ids_cache(user_id, tree)

        return referral_ids

    def set_referral_ids_cache(self, user_id, tree=None):
        referrals = self.get_referrals(user_id, tree)
        referral_ids = [referral.id for referral in referrals.values() if referral is not None]
        cache.set(self.referral_ids_cache_key(user_id), referral_ids, 60*60*3)
        return referral_ids

    def clear_tree_cache(self):
        cache.delete(self.cache_key)

    def _affiliates_sql(self, user_id):
        sql = '''
           WITH RECURSIVE r AS (
              SELECT id, 0 AS level
              FROM users_user
              WHERE id = %s

              UNION ALL

              SELECT users_user.id, r.level + 1 AS level
              FROM users_user
                 JOIN r
                     ON users_user.referral_id = r.id
               WHERE r.level < %s
           )

           SELECT * FROM r;
        '''
        return self.raw(sql, [user_id, settings.AFFILIATE_LEVELS])

    def _referrals_sql(self, user_id):
        sql = '''
           WITH RECURSIVE r AS (
              SELECT id, referral_id, 0 AS level
              FROM users_user
              WHERE id = %s

              UNION ALL

              SELECT users_user.id, users_user.referral_id, r.level + 1 AS level
              FROM users_user
                 INNER JOIN r
                     ON users_user.id = r.referral_id
               WHERE r.level < %s
           )

           SELECT * FROM r;
        '''
        return self.raw(sql, [user_id, settings.AFFILIATE_LEVELS])

    def get_affiliates_from_db(self, user_id):
        levels = {i: [] for i in range(1, settings.AFFILIATE_LEVELS + 1)}
        affiliates_in_use = [user_id]   # prevent cycling
        for user in self._affiliates_sql(user_id):
            if user.level > 0 and user.id not in affiliates_in_use:
                levels[user.level].append(user)
                affiliates_in_use.append(user.id)

        return levels

    def get_referrals_from_db(self, user_id):
        levels = {i: None for i in range(1, settings.AFFILIATE_LEVELS + 1)}
        referrals_in_use = [user_id]    # prevent cycling
        for user in self._referrals_sql(user_id):
            if user.level > 0 and user.id not in referrals_in_use:
                levels[user.level] = user
                referrals_in_use.append(user.id)

        return levels
