import uuid

from django.core.validators import RegexValidator
from django.contrib.auth.models import AbstractUser
from django.utils.translation import ugettext_lazy as _
from django.db.models import (
    CharField, BooleanField, EmailField, ForeignKey, UUIDField, IntegerField,
    PositiveSmallIntegerField, DecimalField, GenericIPAddressField, SET_NULL,
    CASCADE, Sum, Q, DateField, TextField, DateTimeField, OneToOneField,
    PositiveIntegerField, ManyToManyField
)
from django.conf import settings
from django.contrib.postgres.fields import JSONField
from core.constants import LANGUAGE_CHOICES, LANGUAGE_EN
from django_countries.fields import CountryField
from phonenumber_field.modelfields import PhoneNumberField

from user.constants import (
    DEVICE_METHOD, DEVICE_NAME, DEVICE_CHANGE_NAME,
    WAFFLE_HIGH_RISK_LICENSE_ENABLED
)

from user.managers import (
    UserSoftDeletionManager,
    UserTreeManager
)


class User(AbstractUser):
    KYC_NOT_STARTED = 'not_started'
    KYC_INITIATED = 'initiated'
    KYC_DECLINED = 'declined'
    KYC_VERIFIED = 'verified'

    KYC_STATUSES = (
        (KYC_NOT_STARTED, _('Not Started')),
        (KYC_INITIATED, _('Initiated')),
        (KYC_DECLINED, _('Declined')),
        (KYC_VERIFIED, _('Verified')),
    )

    TWO_FACTOR_SMS = 'sms'
    TWO_FACTOR_EMAIL = 'email'
    TWO_FACTOR_AUTHENTICATOR = 'authenticator'
    TWO_FACTOR_DISABLED = 'disabled'

    TWO_FACTOR_CHOICES = (
        (TWO_FACTOR_DISABLED, _('Disabled')),
        (TWO_FACTOR_SMS, _('SMS')),
        # (TWO_FACTOR_EMAIL, 'Email'),
        (TWO_FACTOR_AUTHENTICATOR, _('Authenticator')),
    )

    TYPE_USER = 'user'
    TYPE_COMPANY = 'company'
    USER_TYPES = (
        (TYPE_USER, 'User'),
        (TYPE_COMPANY, 'Company'),
    )

    REGISTER_REFERRAL = 'referral'
    REGISTER_INDEPENDENT = 'independent'
    REGISTER_TYPES = (
        (REGISTER_REFERRAL, 'Referral'),
        (REGISTER_INDEPENDENT, 'Independent')
    )

    uuid = UUIDField(unique=True, default=uuid.uuid4, editable=False)
    # First Name and Last Name do not cover name patterns
    # around the globe.
    name = CharField(
        _("Name of User"),
        blank=True,
        max_length=255,
        validators=[
            RegexValidator(
                regex=r"^[a-zA-Z-_,.' ]+$",
                message=_('Name must contain only letters'),
                code='invalid_name'
            )
        ]
    )
    email = EmailField(_('Email Address'), unique=True)
    phone = PhoneNumberField(_('Phone Number'), null=True, blank=True, unique=True)
    phone_verified = BooleanField(_('Phone is Verified'), default=False)
    kyc_status = CharField(_('KYC'), max_length=16, choices=KYC_STATUSES, default=KYC_NOT_STARTED)
    country = CountryField(blank=True, null=True)
    birthday = DateField(null=True, blank=True)
    toc = BooleanField(_("Terms and Conditions"), default=False)
    referral = ForeignKey(to="self", on_delete=SET_NULL, null=True, blank=True, default=None)
    is_activation_fee_paid = BooleanField(_('Activation fee is paid'), default=False)
    eth_wallet = CharField(_('Ethereum wallet'), max_length=42, blank=True)
    checking_account = CharField(_('Checking account'), max_length=24, null=True, blank=True)
    checking_account_balance_eur = DecimalField(
        _('Checking account balance Eur'), max_digits=10, decimal_places=2, default=0)

    total_licenses = PositiveIntegerField(default=0)
    total_lr_licenses = PositiveIntegerField(default=0)
    total_mr_licenses = PositiveIntegerField(default=0)
    total_hr_licenses = PositiveIntegerField(default=0)

    promo_licenses = PositiveSmallIntegerField(default=0)

    affiliate_licenses = IntegerField(default=0)    # 9 levels of affiliates, used only for counting ranks

    total_lr_eth = DecimalField(
        _('Total Low Eth Price'),
        max_digits=23,
        decimal_places=18,
        default=0
    )
    total_mr_eth = DecimalField(
        _('Total Medium Eth Price'),
        max_digits=23,
        decimal_places=18,
        default=0
    )
    total_hr_eth = DecimalField(
        _('Total High Eth Price'),
        max_digits=23,
        decimal_places=18,
        default=0
    )

    total_lr_usd = DecimalField(_('Total Usd Low'), max_digits=10, decimal_places=2, default=0)
    total_mr_usd = DecimalField(_('Total Usd Medium'), max_digits=10, decimal_places=2, default=0)
    total_hr_usd = DecimalField(_('Total Usd High'), max_digits=10, decimal_places=2, default=0)

    oc_id = IntegerField(default=0, null=True, blank=True)
    bcp_id = IntegerField(default=0, null=True, blank=True)
    wp_no_phone = BooleanField(default=False)
    old_total_licenses = PositiveSmallIntegerField(default=0)
    old_total_usd = DecimalField(_('Old Total Usd'), max_digits=10, decimal_places=2, default=0)
    old_total_btc = DecimalField(_('Old Total BTC'), max_digits=23, decimal_places=18, default=0)
    old_total_eth = DecimalField(_('Old Total Eth'), max_digits=23, decimal_places=18, default=0)

    two_factor_method = CharField(max_length=16, choices=TWO_FACTOR_CHOICES, default=TWO_FACTOR_DISABLED, blank=False)

    rank_info = JSONField(default=dict)     # updates in tasks_residual_bonus.py

    antiphishing_code = CharField(max_length=8, blank=True)
    in_freeze = BooleanField(default=False, help_text=_("When account in freeze, user can't login to the site"))

    new_compensation_plan = BooleanField(default=False)
    notes = TextField(max_length=100, null=True, blank=True)
    user_type = CharField(max_length=8, choices=USER_TYPES, default=TYPE_USER)
    creator = ForeignKey(to="self", related_name='creators', on_delete=SET_NULL, blank=True, null=True)
    register_type = CharField(max_length=12, choices=REGISTER_TYPES, default=REGISTER_REFERRAL)

    deleted_at = DateTimeField(blank=True, null=True)

    language = CharField(max_length=32, choices=LANGUAGE_CHOICES, default=LANGUAGE_EN)

    REQUIRED_FIELDS = ['email', 'phone']

    objects = UserSoftDeletionManager()
    deleted_objects = UserSoftDeletionManager(alive_only=False)
    objects_custom = UserTreeManager()

    def save(self, *args, **kwargs):
        super(User, self).save(*args, **kwargs)
        if self.phone and not self.get_sms_device(names=[DEVICE_NAME, DEVICE_CHANGE_NAME]):
            # creating user sms device for new users
            self.create_sms_device()

    @property
    def phone_masked(self):
        phone = ''
        if self.phone and self.phone != '':
            number = self.phone.as_e164
            phone = f'{number[:6]}...{number[-2:]}'
        return phone

    @property
    def phone_number(self):
        return self.phone

    @property
    def email_masked(self):
        name, domain = self.email.split('@')
        subdomains = domain.split('.')
        return f'{name[:1]}...@{subdomains[0][:1]}...{subdomains[-1]}'

    @property
    def antiphishing_masked(self):
        ac_length = len(self.antiphishing_code)
        code = ''
        if ac_length:
            hl = int(ac_length/2)
            code = f"{self.antiphishing_code[0:hl]}{'*'*(len(self.antiphishing_code)-hl)}"
        return code

    @property
    def is_old_user(self):
        return bool((self.oc_id > 0 or self.bcp_id > 0) and not self.phone_verified)

    def can_invite(self):
        return bool(self.is_active and self.is_activation_fee_paid)

    @property
    def rank(self):
        cache_key = f'rank_v3_{self.id}'
        rank = cache.get(cache_key)
        if not rank:
            rank = self.get_rank()
            cache.set(cache_key, rank, 60*10)
        return rank

    @property
    def is_new_compensation_plan_version(self):
        return True
        # if self.country in settings.NEW_COMPENSATION_PLAN_COUNTRIES:
        #     return True
        # return self.new_compensation_plan

    @property
    def total_usd(self):
        return self.total_lr_usd + self.total_mr_usd + self.total_hr_usd

    @property
    def total_eth(self):
        return self.total_lr_eth + self.total_mr_eth + self.total_hr_eth

    @property
    def daily_payments(self):
        cache_key = f'dp_user_{self.id}'
        dp = cache.get(cache_key)
        if not dp:
            dp = self.paymentorder_set.get_valid_payment_orders().filter(transaction_type='daily_payment').aggregate(
                low=Coalesce(Sum('amount_eth', filter=(Q(daily_income_amount__risk_level=RISK_LOW_CONSTANT))), 0),
                medium=Coalesce(Sum('amount_eth', filter=(Q(daily_income_amount__risk_level=RISK_MEDIUM_CONSTANT))), 0),
                high=Coalesce(Sum('amount_eth', filter=(Q(daily_income_amount__risk_level=RISK_HIGH_CONSTANT))), 0),
                total=Coalesce(Sum('amount_eth'), 0),
            )
            cache.set(cache_key, dp, 60*120)    # cache for 2 hours
        return dp

    @property
    def is_adult(self):
        if self.user_type == self.TYPE_COMPANY:
            return True
        else:
            if not self.birthday:
                return False

            age = now().year - self.birthday.year - ((self.birthday.month, self.birthday.day) > (now().month, now().day))
            if age >= 18:
                return True
            else:
                return False

    def get_rank(self, tree=None):
        if tree is None:
            affiliates = User.objects.filter(referral_id=self.id)
        else:
            all_affiliates = User.objects_custom.get_affiliates(self.id, tree)
            affiliates = all_affiliates[1]

        own_licenses = self.total_licenses

        rank = None
        next_rank = None
        initial_rank = self.rank_info.get('initial_rank', None)
        if initial_rank:
            initial_rank = USER_RANK_VALUE.get(initial_rank, 0)

        for rc in USER_RANKS:
            gift = GIFT_REQUIREMENTS.get(rc['name'], False)
            affiliate_licenses = 0
            first_level_licenses = 0

            """
            Do not apply 40% rule to users who achieved this or higher rank before 40% rule was added
            """
            if initial_rank:
                current_rank = USER_RANK_VALUE.get(rc['name'], 0)
                if current_rank <= initial_rank:
                    gift = False

            for affiliate in affiliates:
                first_level_licenses += affiliate.total_licenses
                if gift:
                    affiliate_licenses += min(affiliate.total_licenses + affiliate.affiliate_licenses, gift['licenses_per_line'])
                else:
                    affiliate_licenses += affiliate.total_licenses + affiliate.affiliate_licenses

            if all([
                own_licenses >= rc['licenses'],
                affiliate_licenses >= rc['licenses_total'],
                first_level_licenses >= rc['licenses_first_level']
            ]):
                rank = rc
            else:
                next_rank = rc
                break

        return dict(rank=rank.get('name', '') if rank else '',
                    user_licenses=own_licenses,
                    affiliate_licenses=affiliate_licenses,
                    first_level_licenses=first_level_licenses,
                    next_rank=next_rank if next_rank else {})

    def update_rank(self, tree=None):
        cache_key = f'rank_v3_{self.id}'
        cache.delete(cache_key)
        rank_info = self.get_rank(tree)
        rank_info['last_updated'] = now().isoformat()
        initial_rank = self.rank_info.get('initial_rank', None)
        if initial_rank and initial_rank == rank_info.get('rank', ''):
            rank_info['initial_rank'] = initial_rank
        self.rank_info = rank_info
        self.save(update_fields=['rank_info'])

    def get_sms_device(self, names=None, confirmed=None, phone=None):
        if phone is None:
            phone = self.phone
        if names is None:
            names = [DEVICE_NAME]
        data = dict(number=phone, method=DEVICE_METHOD, name__in=names)
        if confirmed is not None:
            data['confirmed'] = confirmed
        try:
            device = self.phonedevice_set.get(**data)
        except MultipleObjectsReturned:
            device = self.phonedevice_set.filter(**data).first()
        except ObjectDoesNotExist:
            device = None

        return device

    def create_sms_device(self, name=DEVICE_NAME, confirmed=False, phone=None):
        from two_factor.models import PhoneDevice
        if phone is None:
            phone = self.phone
        data = dict(name=name, method=DEVICE_METHOD, user=self, number=phone, confirmed=confirmed)
        key = random_hex(20).decode('ascii')
        device = PhoneDevice(key=key, **data)
        device.save()

        return device

    def get_or_create_backup_device(self, create=False):
        device, _ = self.staticdevice_set.get_or_create(name='backup')
        if create:
            device.token_set.all().delete()
            self.generate_backup_tokens()
        return device

    def get_email_device(self):
        return self.emaildevice_set.get_or_create(name=DEVICE_NAME, email=self.email)[0]

    def get_authenticator_device(self):
        try:
            device = self.totpdevice_set.get(confirmed=True)
        except MultipleObjectsReturned:
            device = self.totpdevice_set.filter(confirmed=True).order_by('-id').first()
        except ObjectDoesNotExist:
            device = None

        return device

    def get_two_factor_device(self, method=None):
        if method is None:
            method = self.two_factor_method

        if method == self.TWO_FACTOR_SMS:
            return self.get_sms_device(confirmed=True)
        elif method == self.TWO_FACTOR_EMAIL:
            return self.get_email_device()
        elif method == self.TWO_FACTOR_AUTHENTICATOR:
            return self.get_authenticator_device()
        else:
            return None

    @property
    def available_two_factor_methods(self):
        two_factor_methods = [User.TWO_FACTOR_AUTHENTICATOR]

        if self.get_sms_device(confirmed=True):
            two_factor_methods.append(User.TWO_FACTOR_SMS)
        return two_factor_methods

    def generate_backup_tokens(self):
        from django_otp.plugins.otp_static.models import StaticToken
        device = self.get_or_create_backup_device()
        device.token_set.all().delete()
        tokens = []
        for n in range(settings.TWO_FACTOR_RECOVERY_CODES_COUNT):
            token = StaticToken.random_token()
            device.token_set.create(token=token)
            tokens.append(token)

        return tokens

    @property
    def days_left_till_kyc(self):
        till_date = self.date_joined.date() + timedelta(days=settings.CAN_BUY_LICENSES_WITHOUT_KYC_DAYS)
        days_left = (till_date - datetime.today().date()).days
        return days_left

    @property
    def can_buy_licenses_no_kyc(self):
        return self.days_left_till_kyc > 0

    def can_buy_mr_licenses(self, quantity=1):
        # getting all unpaid invoices
        invoices_in_progress = self.invoice_ids.get_unpaid(RISK_MEDIUM_CONSTANT).aggregate(qty=Sum('quantity'))['qty'] or 0
        return (self.total_lr_licenses + self.promo_licenses) - invoices_in_progress - self.total_mr_licenses >= quantity

    def can_buy_hr_licenses(self, quantity=settings.HIGH_RISK_LICENSES_MIN_AMOUNT) -> bool:
        from kuailian.users.helpers import create_fake_user_request
        fake_req = create_fake_user_request(user=self)
        if not waffle.flag_is_active(fake_req, WAFFLE_HIGH_RISK_LICENSE_ENABLED):
            return False

        if quantity < settings.HIGH_RISK_LICENSES_MIN_AMOUNT:
            return False

        if not MifidTest.objects.filter(user_id=self.pk, status=MIFID_APPROVED).exists():
            return False

        # Rule 2-1 ratio: In order to buy HR you require a 2:1 ratio that adds up TIER 1 and/or TIER 2 to buy a TIER 3 product.
        available_licenses = (self.total_mr_licenses + self.total_lr_licenses + self.promo_licenses) // 2

        invoices_in_progress = self.invoice_ids.get_unpaid(RISK_HIGH_CONSTANT).aggregate(qty=Sum('quantity'))['qty'] or 0
        return available_licenses - invoices_in_progress - self.total_hr_licenses >= quantity

    @property
    def is_eu_country(self):
        if self.country is None:
            return False
        return bool(self.country.code in dict(EU_COUNTRIES))

    @property
    def crosschange_kyc_days_left(self):
        till_date = self.date_joined.date() + timedelta(days=settings.KYC_KBANK_READ_ONLY_DAYS)
        days_left = (till_date - datetime.today().date()).days
        return days_left
